/*
 * Project Spitfire/SDL
----------------------------------------------------------------------
	sprite.c - Sprite engine for use with cs.h
----------------------------------------------------------------------
 * © David Olofson, 2001
 */

/*
TODO: Automatic extension of the bank and sprite
TODO: tables as needed when loading banks.
*/

#define	DBG(x)

#include "sprite.h"

#include <stdio.h>
#include <stdlib.h>

#include "SDL.h"
#include "SDL_image.h"


s_blitmodes_t	s_blitmode = S_BLITMODE_NORMAL;
unsigned long	s_colorkey = 0;
unsigned char	s_alpha = SDL_ALPHA_OPAQUE;


/*
----------------------------------------------------------------------
 * Basic Container Management
----------------------------------------------------------------------
 */

static int __alloc_bank_table(s_container_t *c, int count)
{
	c->max = count - 1;
	c->banks = (s_bank_t **)calloc(sizeof(s_bank_t *), count);
	return -(c->banks == NULL);
}

s_container_t *s_new_container(unsigned banks)
{
	s_container_t *ret = (s_container_t *)calloc(sizeof(s_container_t), 1);
	if(!ret)
		return NULL;
	if(__alloc_bank_table(ret, banks) < 0)
		return NULL;
	return ret;
}

void s_delete_container(s_container_t *c)
{
	s_delete_all_banks(c);
	free(c->banks);
	free(c);
}


/*
 * Allocates a new sprite.
 * If the sprite exists already, it's surface will be removed, so
 * that one can safely expect to get an *empty* sprite.
 */
s_sprite_t *s_new_sprite_b(s_bank_t *b, unsigned frame)
{
	s_sprite_t	*s = NULL;
	if(frame > b->max)
		return NULL;
	if(!b->sprites[frame])
	{
		s = calloc(sizeof(s_sprite_t), 1);
		if(!s)
			return NULL;
		b->sprites[frame] = s;
	}
	else
	{
		if(s->surface)
			SDL_FreeSurface(s->surface);
		s->surface = NULL;
	}
	return s;
}


/*
 * Allocates a new sprite.
 * If the sprite exists already, it's surface will be removed, so
 * that one can safely expect to get an *empty* sprite.
 *
 * If the bank does not exists, or if the operation failed for
 * other reasons, this call returns NULL.
 */
s_sprite_t *s_new_sprite(s_container_t *c, unsigned bank, unsigned frame)
{
	if(bank > c->max)
		return NULL;
	if(!c->banks[bank])
		return NULL;

	return s_new_sprite_b(c->banks[bank], frame);
}


void s_delete_sprite_b(s_bank_t *b, unsigned frame)
{
	if(!b->sprites)	/* This can happen when failing to create a new bank */
		return;
	if(frame > b->max)
		return;
	if(!b->sprites[frame])
		return;
	if(b->sprites[frame]->surface)
		SDL_FreeSurface(b->sprites[frame]->surface);
	free(b->sprites[frame]);
	b->sprites[frame] = NULL;
}


void s_delete_sprite(s_container_t *c, unsigned bank, unsigned frame)
{
	s_bank_t *b = s_get_bank(c, bank);
	if(b)
		s_delete_sprite_b(b, frame);
}


static int __alloc_sprite_table(s_bank_t *b, unsigned frames)
{
	b->max = frames - 1;
	b->sprites = (s_sprite_t **)calloc(sizeof(s_sprite_t *), frames);
	if(!b->sprites)
		return -1;
	return 0;
}


s_bank_t *s_new_bank(s_container_t *c, unsigned bank, unsigned frames,
				unsigned w, unsigned h)
{
	s_bank_t *b;
	DBG(printf("s_new_bank(%p, %d, %d, %d, %d)\n", c, bank, frames, w, h));
	if(bank > c->max)
		return NULL;
	if(c->banks[bank])
		s_delete_bank(c, bank);

	c->banks[bank] = (s_bank_t *)calloc(sizeof(s_bank_t), 1);
	if(!c->banks[bank])
		return NULL;
	if(__alloc_sprite_table(c->banks[bank], frames) < 0)
	{
		s_delete_bank(c, bank);
		return NULL;
	}
	b = c->banks[bank];
	b->max = frames - 1;
	b->w = w;
	b->h = h;
	return b;
}


void s_delete_bank(s_container_t *c, unsigned bank)
{
	int i;
	s_bank_t *b;
	if(!c->banks)
		return;
	if(bank > c->max)
		return;
	b = c->banks[bank];
	if(b)
	{
		for(i = 0; i < b->max; ++i)
			s_delete_sprite_b(b, i);
		free(b->sprites);
		free(b);
		c->banks[bank] = NULL;
	}
}


void s_delete_all_banks(s_container_t *c)
{
	int i;
	for(i = 0; i < c->max; ++i)
		s_delete_bank(c, i);
}



/*
----------------------------------------------------------------------
 * Getting data
----------------------------------------------------------------------
 */

s_bank_t *s_get_bank(s_container_t *c, unsigned bank)
{
	if(bank > c->max)
		return NULL;
	if(!c->banks[bank])
		return NULL;
	return c->banks[bank];
}

s_sprite_t *s_get_sprite(s_container_t *c, unsigned bank, unsigned frame)
{
	s_bank_t *b = s_get_bank(c, bank);
	if(!b)
		return NULL;
	if(frame > b->max)
		return NULL;
	return b->sprites[frame];
}

s_sprite_t *s_get_sprite_b(s_bank_t *b, unsigned frame)
{
	if(!b)
		return NULL;
	if(frame > b->max)
		return NULL;
	return b->sprites[frame];
}


/*
----------------------------------------------------------------------
 * File tools
----------------------------------------------------------------------
 */

static int extract_sprite(s_bank_t *bank, unsigned frame,
				SDL_Surface *src, SDL_Rect *from)
{
	SDL_Surface	*tmp;
	SDL_Rect	dr;
	if(frame > bank->max)
	{
		fprintf(stderr, "sprite: Too many frames!\n");
		return -1;
	}
	if(!s_new_sprite_b(bank, frame))
		return -2;

	tmp = SDL_CreateRGBSurface(SDL_SWSURFACE,
			from->w, from->h,
			src->format->BitsPerPixel,
			src->format->Rmask,
			src->format->Gmask,
			src->format->Bmask,
			src->format->Amask	);
	if(!tmp)
		return -3;

	if(src->format->palette)
		SDL_SetColors(tmp, src->format->palette->colors, 0,
				src->format->palette->ncolors);

	dr.x = 0;
	dr.y = 0;
	switch(s_blitmode)
	{
	  case S_BLITMODE_NORMAL:
		if(src->format->Amask)
			SDL_SetAlpha(src, 0, s_alpha);
		SDL_BlitSurface(src, from, tmp, &dr);
		if(src->format->Amask)
			SDL_SetAlpha(tmp, SDL_SRCALPHA | SDL_RLEACCEL,
					SDL_ALPHA_OPAQUE);
		break;
	  case S_BLITMODE_OPAQUE:
		SDL_BlitSurface(src, from, tmp, &dr);
		break;
	  case S_BLITMODE_COLORKEY:
		SDL_BlitSurface(src, from, tmp, &dr);
		SDL_SetColorKey(tmp, SDL_SRCCOLORKEY | SDL_RLEACCEL,
				SDL_MapRGB(tmp->format,
					(s_colorkey>>16) & 0xff,
					(s_colorkey>>8) & 0xff,
					s_colorkey & 0xff)
				);
		break;
	  case S_BLITMODE_ALPHA:
		SDL_BlitSurface(src, from, tmp, &dr);
		SDL_SetAlpha(tmp, SDL_SRCALPHA | SDL_RLEACCEL,
				s_alpha);
		break;
	}
	if(src->format->Amask)
		bank->sprites[frame]->surface = tmp;
	else
	{
		bank->sprites[frame]->surface = SDL_DisplayFormat(tmp);
		SDL_FreeSurface(tmp);
	}
	if(!bank->sprites[frame]->surface)
		return -4;

	DBG(printf("image %d: (%d,%d)/%dx%d @ %p\n", frame,
			from->x, from->y, from->w, from->h,
			bank->sprites[frame]->surface));
	return 0;
}


int s_load_image(s_container_t *c, unsigned bank, const char *name)
{
	SDL_Surface	*src;
	s_bank_t	*b;

	src = IMG_Load(name);
	if(!src)
	{
		fprintf(stderr, "sprite: Failed to load sprite \"%s\"!\n", name);
		return -1;
	}

	b = s_new_bank(c, bank, 1, src->clip_rect.w, src->clip_rect.h);
	if(!b)
	{
		fprintf(stderr, "sprite: Failed to allocate bank for \"%s\"!\n", name);
		return -2;
	}

	if(extract_sprite(b, 0, src, &src->clip_rect) < 0)
	{
		fprintf(stderr, "sprite: Something went wrong while"
				" extracting sprite \"%s\".\n", name);
		return -3;
	}

	SDL_FreeSurface(src);
	return 0;
}


int s_load_sprite(s_container_t *c, unsigned bank, unsigned frame,
					const char *name)
{
	SDL_Surface	*src;
	SDL_Rect	from;
	s_bank_t	*b = s_get_bank(c, bank);
	if(!b)
	{
		fprintf(stderr, "sprite: While loading \"%s\":"
				" Bank %d does not exist!\n", name, bank);
		return -2;
	}

	src = IMG_Load(name);
	if(!src)
	{
		fprintf(stderr, "sprite: Failed to load sprite \"%s\"!\n", name);
		return -1;
	}

	from = src->clip_rect;
	if( (from.w != b->w) || (from.h != b->h) )
	{
		fprintf(stderr, "sprite: Warning: Sprite \"%s\" cropped"
				" to fit in bank %d.\n", name, bank);
		from.w = b->w;
		from.h = b->h;
	}

	if(extract_sprite(b, frame, src, &from) < 0)
	{
		fprintf(stderr, "sprite: Something went wrong while"
				" extracting sprite \"%s\".\n", name);
		return -3;
	}

	SDL_FreeSurface(src);
	return 0;
}


int s_load_bank(s_container_t *c, unsigned bank, unsigned w, unsigned h,
					const char *name)
{
	SDL_Surface	*src;
	s_bank_t	*b;
	int		x, y;
	int		frame = 0;
	int		frames;

	DBG(printf("s_load_bank(%p, %d, %d, %d, %s)\n", c, bank, w, h, name));

	src = IMG_Load(name);
	if(!src)
	{
		fprintf(stderr, "sprite: Failed to load sprite palette %s!\n", name);
		return -1;
	}

	frames = (src->w / w) * (src->h / h);
	b = s_new_bank(c, bank, frames, w, h);
	if(!b)
	{
		fprintf(stderr, "sprite: Failed to allocate bank for \"%s\"!\n", name);
		return -2;
	}

	for(y = 0; y <= src->h - h; y += h)
		for(x = 0; x <= src->w - w; x += w)
		{
			SDL_Rect r;
			r.x = x;
			r.y = y;
			r.w = w;
			r.h = h;
			if(extract_sprite(b, frame, src, &r) < 0)
			{
				fprintf(stderr, "sprite: Something went wrong while"
						" extracting sprites.\n");
				return -3;
			}
			++frame;
		}

	SDL_FreeSurface(src);
	return 0;
}
